/*
 ************************************************************************
 *                                                                      *
 *      UART.C                                                          *
 *                                                                      *
 *      Buffered serial IO under DOS                                    *
 *                                                                      *
 ************************************************************************

 ************************************************************************
 * Supported by :                                                       *
 *              Rolf Segger * Wilhelmine-Fliedner-Str. 9                *
 *              40723 Hilden * Germany                                  *
 *              Tel 49-2103-62915 * Fax 49-2103-62916                   *
 *                                                                      *
 ************************************************************************
*/



#include <stdlib.h>
#include <dos.h>
#include "UART.H"

#define SetVect(x,y) setvect(x,y)
#define GetVect(x)   getvect(x)
#define Enable()     enable()
#define Disable()    disable()

//	****************************************
//	*                                      *
//	*  Serial interupt data & declarations *
//	*                                      *
//	****************************************

typedef void interrupt (PINTHANDLER)(...);
PINTHANDLER	IntHandlerCom;
PINTHANDLER* pSerIntOld;

int BasePort[] = { 0x3f8, 0x2f8 };
int IntMask[] = { 0x10, 0x8 };
int Intr[] = { 0x0c, 0x0b };


//	*********************************
//	*				*
//	*	Variables		*
//	*				*
//	*********************************

char TXSBusy=0;
char Cid=-1;

//	******************************************************
//	*				                     *
//	*     Buffer-management	                             *
//	*				                     *
//	******************************************************

#define BUFSIZE	(512)
static char WrBuf[BUFSIZE];
int RdOff, WrOff;

int UARTGetTxCnt() {
  disable();
  int Cnt = WrOff - RdOff;
  if (Cnt < 0) Cnt+= BUFSIZE;
  enable();
  return Cnt;
}

char UARTStoreTx(char c) {
  if (UARTGetTxCnt()==BUFSIZE-1) return 1;
  WrBuf[WrOff++] = c;
  if (WrOff==BUFSIZE) WrOff = 0;
  return 0;
}

char UARTGet() {
  char c = WrBuf[RdOff++];
  if (RdOff==BUFSIZE) RdOff = 0;
  return c;
}

//	******************************************************
//	*				                     *
//	*     UART-routines   	                             *
//	*				                     *
//	*     for interrupt-driven communication             *
//	*				                     *
//	******************************************************

void UARTWriteChar(char c) {
  TXSBusy=1;
  outp (BasePort[Cid], c);
}

char UARTWriteCharBuffer(char c) {
  disable();
    if (!TXSBusy) goto Write;
  enable();
  return UARTStoreTx(c);
Write:
  enable();
  UARTWriteChar(c);
  return 0; 		// O.K., no problem !
}

void UARTReadChar() { UARTStoreRxdChar(inp(BasePort[Cid])); }

int wc;
void UARTWriteNext() {
  wc++;
  if (UARTGetTxCnt()) UARTWriteChar(UARTGet());
  else TXSBusy = 0;
}

PINTHANDLER	IntHandlerCom {
  char IntId = inp(BasePort[Cid]+2);		// interrupt identification register
  if (IntId == 4) UARTReadChar();
  if (IntId == 2) UARTWriteNext();
  outp (0x20,0x20);
}


//	******************************************************
//	*				                     *
//	*     UART-init   	                             *
//	*				                     *
//	******************************************************



void UARTInitHard(int cid, int baud) {
  unsigned cbaud = (115200L / baud);
  char c;
  Cid = cid;
  outp(BasePort[cid]+3, 0x83);		// set DLAB
  outp(BasePort[cid],   cbaud&0xff);
  cbaud = cbaud>>8;
  outp(BasePort[cid]+1, cbaud);
  outp(BasePort[cid]+3, 3);		// clear DLAB, n,8,1
}

int UARTSetInt(int cid) {
  if (inp(BasePort[cid])==0xFF && inp(BasePort[cid]+5)==0xFF && inp(BasePort[cid]+3)==0xFF) return -1;   /* no UART found at port address */
  pSerIntOld = GetVect(Intr[cid]);	// save old handler
  SetVect(Intr[cid], IntHandlerCom); // point to our handler
  disable();
  outp( BasePort[cid]+1, 3);		// set interupt enable Rx/Tx
  outp( BasePort[cid]+4, 0xb);   	// set OUT2, DTR, & RTS on UART
  enable();
  outp(0x21, inp(0x21) & ~IntMask[cid]);	// setup interupt controller chip
  return 0;
}

void UARTInit(char cid, long baudrate) {
  UARTInitHard(cid, baudrate);
  UARTSetInt( cid);
}

//	******************************************************
//	*				                     *
//	*     UART-exit-routines                             *
//	*				                     *
//	******************************************************

void UARTResetInt(void) {
  if (Cid>=0) {
    outp (0x21, inp(0x21) | IntMask[Cid]);
    SetVect(Intr[Cid], pSerIntOld);
  }
}

void UARTExit() {
  outp( BasePort[Cid]+4, 0x0);   	// reset OUT2, DTR, & RTS on UART
  UARTResetInt();
}
